import hashlib
import json
import logging
import os
import traceback
from datetime import datetime, date, time
from datetime import timedelta
from decimal import Decimal
from enum import Enum
from pathlib import Path
from typing import Any
from uuid import UUID

import sys
from slack_sdk import WebClient as SlackWebClient


def safe_serialize(obj):
    if isinstance(obj, (datetime, date, time)):
        return obj.isoformat()
    if isinstance(obj, (set, frozenset)):
        return list(obj)
    if isinstance(obj, (bytes, bytearray)):
        try:
            return obj.decode()
        except Exception:
            return str(obj)
    if isinstance(obj, (Decimal, UUID, Path, Enum, complex)):
        return str(obj)
    if isinstance(obj, Exception):
        return str(obj)
    if hasattr(obj, '__dict__'):
        return vars(obj)
    return str(obj)


class SlackLogLevel(str, Enum):
    INFO = "good"
    WARNING = "warning"
    ERROR = "danger"
    DEBUG = "#CCCCCC"  # light gray

    @property
    def prefix(self) -> str:
        if self == SlackLogLevel.DEBUG:
            return "*DEBUG*: "
        elif self == SlackLogLevel.WARNING:
            return "*WARNING*: "
        return ""


class BotIconEmoji(str, Enum):
    """
    Enum for Slack icon emojis.
    The SLACK_LOGGER_EMOJI is a reference to an environment variable
        that can be set to customize the emoji used by the Slack logger.
        If not set, it will default to `:robot_face:`
    """
    TECHNOLOGIST = ":technologist:"
    ROBOT_FACE = ":robot_face:"
    SLACK_LOGGER_EMOJI = "SLACK_LOGGER_EMOJI"  # symbolic reference, value will be os.getenv("SLACK_LOGGER_EMOJI")


class SlackLogger:
    def __init__(
            self,
            project_name: str,
            slack_token: str,
            subprocess: str = None,
            add_subprocess_to_info: bool = True,
            default_channel_id: str = None,
            info_channel_id: str = None,
            error_channel_id: str = None,
            icon_emoji: str | BotIconEmoji = None,
    ) -> None:
        """
        Initialize the SlackLogger instance.

        :param project_name: name of the project for logging purposes (used in the username of the slack bot)
        :param slack_token: Slack API token to authenticate the Slack client.
        :param subprocess: (Optional) Name of the subprocess for logging context. This will be prefixed to messages.
        :param add_subprocess_to_info: Whether to include the subprocess name in info messages. Default to True.
        :param default_channel_id: default Slack channel ID to send messages to if no specific channel is provided.
        :param info_channel_id: default Slack channel ID for info messages.
        :param error_channel_id: default Slack channel ID for error messages.
        :param icon_emoji:
            - str (custom emoji), e.g. ":robot_face:"
            - BotIconEmoji Enum member, e.g. BotIconEmoji.TECHNOLOGIST
            If neither is provided, defaults to:
              - BotIconEmoji.SLACK_LOGGER_EMOJI (if set via environment),
              - otherwise BotIconEmoji.TECHNOLOGIST.
        :raises ValueError: If no channel IDs are provided.
        """
        if not any([default_channel_id, info_channel_id, error_channel_id]):
            raise ValueError("At least one channel ID must be provided.")

        self._project_name = project_name
        self._slack_client = SlackWebClient(slack_token)

        self._default_channel_id = default_channel_id or info_channel_id or error_channel_id
        self._info_channel_id = info_channel_id
        self._error_channel_id = error_channel_id

        self._provided_icon_emoji = icon_emoji
        self._icon_emoji = self._get_bot_emoji()

        self._last_messages = []
        self._subprocess = subprocess
        self._add_subprocess_to_info = add_subprocess_to_info

    def _get_bot_emoji(self) -> str:
        """
        Determines the emoji to use for the Slack bot.
        if icon_emoji is set to BotIconEmoji.SLACK_LOGGER_EMOJI,
            it will check for the environment variable `SLACK_LOGGER_EMOJI`
            if not set, it will fallback to the default `:robot_face:`.
        :return: str representation of the emoji to use for the Slack bot.
        """
        env_emoji = os.getenv("SLACK_LOGGER_EMOJI")

        if self._provided_icon_emoji == BotIconEmoji.SLACK_LOGGER_EMOJI:
            if env_emoji:
                return env_emoji
            logging.warning("Warning: SLACK_LOGGER_EMOJI environment variable is not set, "
                            "falling back to default icon :robot_face:")
            return BotIconEmoji.ROBOT_FACE.value

        if self._provided_icon_emoji:
            if isinstance(self._provided_icon_emoji, BotIconEmoji):
                self._provided_icon_emoji = self._provided_icon_emoji.value
            return self._provided_icon_emoji

        return env_emoji or BotIconEmoji.TECHNOLOGIST.value

    def clone(self,
              *,
              subprocess: str = None,
              add_subprocess_to_info: bool = True,
              default_channel_id: str = None,
              info_channel_id: str = None,
              error_channel_id: str = None,
              icon_emoji: str | BotIconEmoji = None
              ) -> "SlackLogger":
        """
        Clone the current SlackLogger instance, allowing for different configurations.
        """
        return SlackLogger(
            project_name=self._project_name,
            slack_token=self._slack_client.token,
            subprocess=subprocess or self._subprocess,
            add_subprocess_to_info=add_subprocess_to_info if add_subprocess_to_info is not None else self._add_subprocess_to_info,
            default_channel_id=default_channel_id or self._default_channel_id,
            info_channel_id=info_channel_id or self._info_channel_id,
            error_channel_id=error_channel_id or self._error_channel_id,
            icon_emoji=icon_emoji or self._icon_emoji
        )

    def _resolve_channel(self, provided_channel_id: str, is_error: bool) -> str:
        """
        Resolve the channel ID to use for posting messages.
        """
        if is_error:
            return provided_channel_id or self._error_channel_id or self._default_channel_id
        return provided_channel_id or self._info_channel_id or self._default_channel_id

    def _hash_error(self, error_text: str) -> str:
        """
        Generate a hash for the error text to avoid duplicate messages.
        """
        return hashlib.md5(error_text.encode()).hexdigest()

    @staticmethod
    def _serialize_context_data(context_data: Any) -> str:
        """
        Serialize context data to a string format for logging.
        """
        if isinstance(context_data, str):
            return context_data

        return json.dumps(context_data, indent=4, default=safe_serialize, ensure_ascii=False)

    def _post_to_slack(self,
                       message: str,
                       level: SlackLogLevel,
                       subprocess: str = None,
                       error_text: str = None,
                       channel_id: str = None,
                       user_name: str = None,
                       color: str = None) -> None:
        """ Post ordinary messages as warning or error messages as danger
        :param message: message to post
        :param level: message level (error, warning, info)
        :param subprocess: Optional subprocess name to include in the message.
        :param error_text: error text to post
        :param channel_id: Slack channel ID to send the message to, if different from the default
        :param user_name: Optional username to use for the message.
        :param color: Optional HEX or Slack-supported color ('good', 'warning', 'danger').
        """

        message = level.prefix + message if message else level.prefix

        subprocess_name = subprocess or self._subprocess

        if level == SlackLogLevel.INFO and not self._add_subprocess_to_info:
            """
            Do not include subprocess in info messages unless explicitly included in the constructor.
            This allows cleaner info messages without subprocess context.
            """
            subprocess_name = None

        message = f"[{subprocess_name}]: {message}" if subprocess_name else message

        attachments = {
            "text": message,
            "fallback": message,
            "color": color or level.value
        }

        if level == SlackLogLevel.ERROR and error_text:
            attachments.update({
                "text": error_text,
                "pretext": message,
                "fallback": message,
                "title": "Error traceback"
            })

        self._slack_client.chat_postMessage(
            channel=self._resolve_channel(channel_id, is_error=level == SlackLogLevel.ERROR),
            attachments=[attachments],
            username=user_name or self._project_name.lower(),
            icon_emoji=self._icon_emoji
        )

    def _write_error_log_and_post(self,
                                  error_text: str,
                                  message: str,
                                  subprocess: str = None,
                                  channel_id: str = None,
                                  user_name: str = None,
                                  color: str = None):
        self._last_messages.append({
            'date': datetime.now().replace(microsecond=0),
            'message_hash': self._hash_error(error_text)
        })

        if not os.path.exists("logs"):
            os.makedirs("logs")

        with open(f"logs/{self._project_name}.log", "w", encoding='utf-8') as file:
            file.write(f'\n\nDATE: {datetime.now().replace(microsecond=0)}: ERROR in {message}\n\n')
            e_type, e_val, e_tb = sys.exc_info()
            traceback.print_exception(e_type, e_val, e_tb, file=file)

        truncated = error_text[-8000:] if len(error_text) > 8000 else error_text
        self._post_to_slack(
            message=message,
            level=SlackLogLevel.ERROR,
            subprocess=subprocess,
            error_text=truncated,
            channel_id=channel_id,
            user_name=user_name,
            color=color
        )

    def error(self,
              exc: Exception,
              subprocess: str = None,
              header_message: str = None,
              context_data: Any = None,
              channel_id: str = None,
              user_name: str = None,
              color: str = None) -> None:
        """
        :param exc: Exception object appears as red text in slack.
        :param subprocess: Optional subprocess name to include in the message.
        :param header_message: bold text appears above the error message - usually the place where the error occurred
        :param context_data: Additional data to be added to the error message like variables, API json payload, etc.
        :param channel_id: Slack channel ID to send the message to, if different from the default
        :param user_name: Optional username to use for the message.
        :param color: Optional HEX or Slack-supported color ('good', 'warning', 'danger').
        :return: None
        """
        error_text = ''.join(traceback.format_exception(None, exc, exc.__traceback__))
        if context_data:
            context_data = self._serialize_context_data(context_data)

            if header_message:
                header_message += f"\nContext data:\n{context_data}"
            else:
                header_message = f"Context data:\n{context_data}"

        error_hash = self._hash_error(error_text)
        two_minutes_ago = datetime.now() - timedelta(minutes=2)
        self._last_messages = [
            item for item in self._last_messages if item['date'] > two_minutes_ago
        ]
        if error_hash in [item['message_hash'] for item in self._last_messages]:
            return

        self._write_error_log_and_post(
            error_text=error_text,
            message=header_message,
            subprocess=subprocess,
            channel_id=channel_id,
            user_name=user_name,
            color=color
        )

    def info(self,
             message: str,
             subprocess: str = None,
             context_data: Any = None,
             channel_id: str = None,
             user_name: str = None,
             color: str = None) -> None:
        """
        Post an info message to Slack with green color.
        :param message: message appears as an info message in slack without error
        :param subprocess: Optional subprocess name to include in the message.
        :param context_data: Additional data to be added to the info message like variables, API json payload, etc.
        :param channel_id: Slack channel ID to send the message to, if different from the default
        :param user_name: Optional username to use for the message.
        :param color: Optional HEX or Slack-supported color ('good', 'warning', 'danger').
        :return: None
        """
        if context_data:
            context_data = self._serialize_context_data(context_data)
            message += f"\n\nContext data:\n{context_data}"

        self._post_to_slack(
            message=message,
            level=SlackLogLevel.INFO,
            subprocess=subprocess,
            channel_id=channel_id,
            user_name=user_name,
            color=color
        )

    def warning(self,
                message: str,
                subprocess: str = None,
                context_data: Any = None,
                channel_id: str = None,
                user_name: str = None,
                color: str = None) -> None:
        """
        Post a warning message to Slack with yellow color.
        :param message: message appears as a warning message in slack without error
        :param subprocess: Optional subprocess name to include in the message.
        :param context_data: Additional data to be added to the warning message like variables, API json payload, etc.
        :param channel_id: Slack channel ID to send the message to, if different from the default
        :param user_name: Optional username to use for the message.
        :param color: Optional HEX or Slack-supported color ('good', 'warning', 'danger').
        :return: None
        """
        if context_data:
            context_data = self._serialize_context_data(context_data)
            message += f"\n\nContext data:\n{context_data}"
        self._post_to_slack(
            message=message,
            level=SlackLogLevel.WARNING,
            subprocess=subprocess,
            channel_id=channel_id,
            user_name=user_name,
            color=color
        )

    def debug(self,
              message: str,
              subprocess: str = None,
              context_data: Any = None,
              channel_id: str = None,
              user_name: str = None,
              color: str = None) -> None:
        """
        Post a debug message to Slack with gray color.
        :param message: message appears as a debug message in slack without error
        :param subprocess: Optional subprocess name to include in the message.
        :param context_data: Additional data to be added to the debug message like variables, API json payload, etc.
        :param channel_id: Slack channel ID to send the message to, if different from the default
        :param user_name: Optional username to use for the message.
        :param color: Optional HEX or Slack-supported color ('good', 'warning', 'danger').
        :return: None
        """
        if context_data:
            context_data = self._serialize_context_data(context_data)
            message += f"\n\nContext data:\n{context_data}"
        self._post_to_slack(
            message=message,
            level=SlackLogLevel.DEBUG,
            subprocess=subprocess,
            channel_id=channel_id,
            user_name=user_name,
            color=color
        )



from redis import Redis
from redis.connection import ConnectionPool
from utils_b_infra.caching.backends.base import BaseCache


class RedisCache(BaseCache):
    def __init__(self,
                 url: str = None,
                 host: str = None,
                 port: int = None,
                 password: str = None,
                 db: int = 0,
                 app_space: str | None = "",
                 default_timeout: int | None = 300) -> None:

        if not url and not host:
            raise ValueError("Either url or host, port and password must be provided")
        if url:
            self._redis = Redis(connection_pool=ConnectionPool.from_url(url=url))
        else:
            self._redis = Redis(connection_pool=ConnectionPool(host=host, port=port, password=password, db=db))
        self._app_space = app_space
        self._default_timeout = default_timeout

    def get_with_ttl(self, key: str) -> tuple[int, bytes | None]:
        with self._redis.pipeline() as pipe:
            ttl, value = pipe.ttl(key).get(key).execute()
            return ttl, value

    def get(self, key: str) -> bytes | None:
        return self._redis.get(key)

    def set(self, key: str, value: bytes, expire: int = None) -> None:
        """
        Set expire to 0 to make it never expire
        """
        self._redis.set(key, value, ex=expire if expire is not None else self._default_timeout)

    def clear(self, namespace: str = None, key: str = None) -> int:
        if key:
            return self._redis.delete(key)

        pattern = ""
        if self._app_space:
            pattern += f"{self._app_space}:"
        if namespace:
            pattern += f"{namespace}:"
        pattern += "*"

        lua = f"""
        local keys = redis.call('KEYS', '{pattern}')
        for i = 1, #keys, 5000 do
            redis.call('DEL', unpack(keys, i, math.min(i + 4999, #keys)))
        end
        return #keys
        """
        return self._redis.eval(lua, numkeys=0)



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()
    long_description = long_description.split('## Changelog')[0]
with open('CHANGELOG.md', 'r') as ch:
    long_description += "\n\n" + ch.read()

setup(
    name="utils_b_infra",
    version="1.3.22",
    author="Fahad Mawlood",
    author_email="fahadukr@gmail.com",
    description="A collection of utility functions and classes for Python projects.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Fahadukr/utils-b-infra",
    packages=find_packages(),
    install_requires=[
        "SQLAlchemy",
        "pandas",
        "numpy < 2.0.0",
        "openai >= 1.74.0",
        "redis",
        "pymongo",
        "pendulum",
        "slack-sdk",
        "tiktoken",
        "google-api-python-client",
        "google-auth-httplib2",
        "google-auth-oauthlib",
        "Pillow",
        "pdf2image",
        "pydub"
    ],
    extras_require={
        'translation': [
            "google-cloud-translate >= 3.12.0",
            "deepl"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
)



import ast
import base64
import io
import json
import os
import re
from typing import List, Union, Any, Literal

import openai
import requests
import tiktoken
from PIL import Image
from openai._types import NOT_GIVEN
from pdf2image import convert_from_bytes
from pydub import AudioSegment
from utils_b_infra.generic import retry_with_timeout


def count_tokens_per_text(text: str) -> int:
    encoding = tiktoken.get_encoding("cl100k_base")
    tokens = encoding.encode(text)
    tokens = len(tokens)
    return tokens


def calculate_openai_price(text: str, output_tokens: int, model: str) -> float:
    """
    Calculate the price for the OpenAI API based on the input text and the output tokens.
    :return: The total price in USD.
    """
    # Input token counts
    input_token_count = count_tokens_per_text(text)

    # Price (USD) per model per 1M tokens (input, output)
    prices = {
        'gpt-4.5-preview': (75, 150),
        'gpt-4o': (2.5, 10),
        'gpt-4o-audio-preview': (2.5, 10),
        'gpt-4o-realtime-preview': (5, 20),
        'gpt-4o-mini': (0.15, 0.60),
        'gpt-4o-mini-audio-preview': (0.15, 0.60),
        'gpt-4o-mini-realtime-preview': (0.60, 2.40),
        'o1': (15, 60),
        'o1-pro': (150, 600),
        'o3-mini': (1.1, 4.4),
        'o1-mini': (1.1, 4.4),
        'gpt-4o-mini-search-preview': (0.15, 0.60),
        'gpt-4o-search-preview': (5.00, 20.00),
    }

    # Calculate price
    price_per_million_input, price_per_million_output = prices[model]
    total_price = ((input_token_count * price_per_million_input) + (
            output_tokens * price_per_million_output)) / 1_000_000

    return total_price


def extract_json_from_text(text_):
    # Regular expression to match the outermost curly braces and their contents
    match = re.search(r'\{.*\}', text_, re.DOTALL)
    if match:
        return match.group(0)
    return None


class TextGenerator:
    def __init__(self, openai_client: openai.Client):
        self.openai_client = openai_client

    # -------------- TEXT-ONLY FUNCTIONALITY --------------

    @staticmethod
    def _parse_json(ai_text):
        for parser in (json.loads, ast.literal_eval):
            try:
                return parser(ai_text)
            except Exception as e:
                last_exception = e
                continue
        print('Error loading JSON from AI text:', last_exception)
        raise ValueError('Invalid JSON format') from last_exception

    @retry_with_timeout(retries=3, timeout=60, initial_delay=10, backoff=2)
    def generate_text_embeddings(self, content, model="text-embedding-3-small"):
        content = content.encode(encoding='ASCII', errors='ignore').decode()
        content = content.replace("\n", " ")
        emb = self.openai_client.embeddings.create(input=content, model=model)
        return emb.data[0].embedding

    @retry_with_timeout(retries=3, timeout=200, initial_delay=10, backoff=2)
    def generate_ai_response(
            self,
            prompt: str,
            user_text: Any = None,
            gpt_model: str = 'gpt-5',
            max_output_tokens: int = NOT_GIVEN,
            temperature: float = 0.5,
            verbosity: Literal["low", "medium", "high"] = "medium",
            reasoning_effort: Literal["minimal", "low", "medium", "high"] = "medium",
            store: bool = False,
            json_mode: bool = False
    ) -> dict | str:
        """
        Generate AI response for the provided user text.
        To process file or audio, use process_file or transcribe_audio_file.

        models 5 and o require reasoning_effort and verbosity parameters,
            and they don't support temperature and max_output_tokens.
        models 4.x support temperature and max_output_tokens,
            but don't support reasoning_effort and verbosity parameters.

        :param prompt: Prompt to be used for the AI model
        :param user_text: Text or JSON object to be used as input
        :param gpt_model: Model to be used for the AI response
        :param max_output_tokens: Max output tokens
        :param temperature: Temperature for the AI model
        :param verbosity: Verbosity level for reasoning models
        :param reasoning_effort: Reasoning effort for reasoning models
        :param store: If True, the response will be stored in the OpenAI API
        :param json_mode: If True, the response will be in JSON format
        :return: AI response as a string or JSON object
        """

        if user_text and not isinstance(user_text, str):
            user_text = json.dumps(user_text)

        # Build the new "input" list
        input_list = [{
            "role": "developer" if gpt_model.startswith('gpt-5') or gpt_model.startswith('o') else "system",
            "content": [
                {"type": "input_text", "text": prompt}
            ]
        }]
        if user_text:
            input_list.append({
                "role": "user",
                "content": [
                    {"type": "input_text", "text": user_text}
                ]
            })

        # Output format for JSON mode:
        request_kwargs = {
            "text": {
                "format": {
                    "type": "json_object" if json_mode else "text"
                },
            }
        }
        if gpt_model.startswith('gpt-5'):
            request_kwargs["text"]["verbosity"] = verbosity

        if gpt_model.startswith('gpt-5') or gpt_model.startswith('o'):
            # Reasoning models
            ai_resp = self.openai_client.responses.create(
                model=gpt_model,
                input=input_list,
                reasoning={"effort": reasoning_effort},
                store=store,
                **request_kwargs
            )
        else:
            # gpt 4 models
            ai_resp = self.openai_client.responses.create(
                model=gpt_model,
                input=input_list,
                temperature=temperature,
                max_output_tokens=max_output_tokens,
                **request_kwargs
            )

        ai_text = ai_resp.output_text

        if ai_text and json_mode:
            ai_text = self._parse_json(ai_text)

        ai_text = "" if ai_text == "''" else ai_text

        return ai_text

    # -------------- IMAGE & FILE HANDLING FUNCTIONALITY --------------

    @staticmethod
    def _get_file_extension(path: str) -> str:
        return os.path.splitext(path)[-1].lstrip(".").lower()

    @staticmethod
    def _download_file_into_bytes(url: str) -> Union[io.BytesIO, None]:
        response = requests.get(url)
        if response.status_code == 200:
            return io.BytesIO(response.content)
        print(f"Failed to download file. Status code: {response.status_code}")
        return None

    @staticmethod
    def _load_local_file_into_bytes(file_path: str) -> Union[io.BytesIO, None]:
        try:
            with open(file_path, "rb") as f:
                return io.BytesIO(f.read())
        except Exception as e:
            print(f"Failed to read local file: {e}")
            return None

    @staticmethod
    def _pdf_page_to_image_from_bytes(pdf_data: Union[bytes, io.BytesIO]) -> List[Image.Image]:
        if isinstance(pdf_data, io.BytesIO):
            pdf_data = pdf_data.getvalue()
        return convert_from_bytes(pdf_data)

    @staticmethod
    def _encode_image(image: Image.Image, width: int = None, height: int = None) -> str:
        if width and height:
            image = image.resize((width, height), Image.ANTIALIAS)
        buffered = io.BytesIO()
        image.save(buffered, format="PNG")
        return base64.b64encode(buffered.getvalue()).decode("utf-8")

    def _build_image_prompt(self, images: List[Image.Image]) -> dict:
        content_items = [{"type": "image_url", "image_url": {
            "url": f"data:image/png;base64,{self._encode_image(img)}"}} for img in images]

        return {"role": "user", "content": content_items}

    def _get_image_gpt_response(self,
                                model: str,
                                system_prompt: str,
                                images: List[Image.Image],
                                temperature: float = 0.2,
                                json_mode: bool = False
                                ) -> dict | str:

        messages = [{"role": "system", "content": system_prompt}]
        messages.append(self._build_image_prompt(images))

        gpt_answer = self.openai_client.chat.completions.create(
            model=model,
            messages=messages,
            response_format={"type": "json_object"} if json_mode else NOT_GIVEN,
            temperature=temperature
        )

        ai_text = gpt_answer.choices[0].message.content
        if ai_text and json_mode:
            try:
                ai_text = json.loads(ai_text, strict=False)
            except Exception as e:
                print('error loading json with json.loads')
                try:
                    ai_text = ast.literal_eval(ai_text)
                except Exception as e:
                    print('error loading json with ast.literal_eval')
                    raise e
        return ai_text

    def process_file(self,
                     prompt: str,
                     model: str = 'gpt-4.1',
                     url: str = None,
                     file_path: str = None,
                     temperature: float = 0.2,
                     json_mode: bool = False
                     ) -> dict | str:
        """
        Process text files like .pdf, .docx, .txt, etc. from a URL or local file.
        Provide either `url` or `file_path`.
        if both are provided, `url` will be used.
        """
        if url:
            ext = self._get_file_extension(url)
            file_data = self._download_file_into_bytes(url)
        elif file_path:
            ext = self._get_file_extension(file_path)
            file_data = self._load_local_file_into_bytes(file_path)
        else:
            raise ValueError("Either 'url' or 'file_path' must be provided.")

        if not file_data:
            return {}

        if ext == 'pdf':
            images = self._pdf_page_to_image_from_bytes(file_data)
        else:
            images = [Image.open(file_data).convert("RGB")]

        return self._get_image_gpt_response(
            model=model,
            system_prompt=prompt,
            images=images,
            temperature=temperature,
            json_mode=json_mode
        )

    def transcribe_audio_file(self,
                              url: str = None,
                              file_path: str = None,
                              model: str = "gpt-4o-transcribe",
                              prompt: str = None,
                              ) -> str:
        """
        Transcribe an audio file (e.g., .oga, .mp3, .wav) to text using OpenAI's transcription model.
        Supports input from a URL or a local file. If both are provided, URL takes precedence.
        Converts to mp3 only if format is not directly supported by OpenAI.

        You can use a prompt to improve the quality of the transcripts generated by the Transcriptions
        example: "The following conversation is a lecture about the recent developments around OpenAI, GPT-4.5 and the future of AI."
        """
        if url:
            response = requests.get(url)
            if response.status_code != 200:
                raise Exception(f"Failed to download audio file from URL. Status code: {response.status_code}")
            audio_bytes = io.BytesIO(response.content)
            audio_format = self._get_file_extension(url)
        elif file_path:
            with open(file_path, "rb") as f:
                audio_bytes = io.BytesIO(f.read())
            audio_format = self._get_file_extension(file_path)
        else:
            raise ValueError("Either 'url' or 'file_path' must be provided.")

        # Clean and normalize extension
        audio_format = audio_format.lower()
        audio_format = 'ogg' if audio_format == 'oga' else audio_format

        if audio_format not in {"mp3", "mp4", "mpeg", "mpga", "m4a", "wav", "webm"}:  # Supported formats by OpenAI
            # Convert to MP3 using pydub if not supported
            audio = AudioSegment.from_file(audio_bytes, format=audio_format)
            audio_bytes = io.BytesIO()
            audio.export(audio_bytes, format="mp3")
            audio_bytes.seek(0)
            audio_bytes.name = "converted.mp3"
        else:
            # Ensure the buffer is ready for reading
            audio_bytes.seek(0)
            audio_bytes.name = f"audio.{audio_format}"

        transcription = self.openai_client.audio.transcriptions.create(
            model=model,
            file=audio_bytes,
            response_format="text",
            prompt=prompt if prompt else NOT_GIVEN,
        )

        return transcription
